@@ -99,23 +99,27 @@ async def query(ctx, query, results):
         sub = storage.subscription_class(
             storage, "cli", query, queue=queue, default_limit=60000
         )
-        sub.prepare()
+        if sub.prepare():
+            await sub.run_query()
+        else:
+            results = False
         click.echo(click.style("Query:", bold=True))
         click.echo(click.style(sub.query, fg="green"))
-        await sub.run_query()
 
-    if results:
-        click.echo(click.style("Results:", fg="black", bold=True))
+        if results:
+            click.echo(click.style("Results:", fg="black", bold=True))
 
-        while True:
-            sub, event = await queue.get()
-            if event:
-                click.echo(
-                    click.style(json.dumps(event.to_json_object(), indent=4), fg="red")
-                )
-                click.echo("")
-            else:
-                break
+            while True:
+                sub, event = await queue.get()
+                if event:
+                    click.echo(
+                        click.style(
+                            json.dumps(event.to_json_object(), indent=4), fg="red"
+                        )
+                    )
+                    click.echo("")
+                else:
+                    break
 
 
 @main.command()
@@ -303,7 +307,7 @@ async def set(pubkey, roles):
         return -1
 
     async with get_storage() as storage:
-        await storage.authenticator.set_roles(pubkey, roles)
+        await storage.set_auth_roles(pubkey, roles)
         click.echo(f"Set roles for pubkey: {pubkey} to {roles}")
 
 
@@ -318,10 +322,10 @@ async def get(pubkey):
 
     async with get_storage() as storage:
         if not pubkey:
-            async for pubkey, role in storage.authenticator.get_all_roles():
+            async for pubkey, role in storage.get_all_auth_roles():
                 click.echo(f"{pubkey}: {role}")
         else:
-            roles = await storage.authenticator.get_roles(pubkey)
+            roles = await storage.get_auth_roles(pubkey)
             click.echo(roles)
 
 
